""" The Pylenium CLI.

We are using `click` to create the CLI interface.
In short, the structure of the interface looks like this:

@click.group() as cli         # this is then publicly exposed as `pylenium` in the user's terminal
    @cli.command()            # A
    @cli.command()            # B

    @cli.group() as portal
        @portal.command()     # C
        @portal.command()     # D

This structure is what gives us commands and sub-commands.
Examples:
    # Use command A
    $ pylenium init

    # Use portal command C
    $ pylenium portal download

For more information, visit their official docs: https://click.palletsprojects.com/en/7.x/
"""

import os
import shutil

import click
from pyfiglet import Figlet

from pylenium.scripts import report_portal


def _copy(file, to_dir, message) -> str:
    """ Copies a file to the given directory.

    If the file exists, it is overwritten.

    Returns:
        The absolute path of the newly copied file.
    """
    newly_created_path = shutil.copy(src=file, dst=to_dir)
    click.echo(f'{message} {newly_created_path}')
    return newly_created_path


@click.group()
@click.version_option()
def cli():
    """ The Pylenium CLI. """
    pass


@cli.command()
@click.option('-c', '--overwrite-conftest', type=bool, show_default=True, is_flag=True)
@click.option('-p', '--overwrite-pylenium-json', type=bool, show_default=True, is_flag=True)
@click.option('-i', '--overwrite-pytest-ini', type=bool, show_default=True, is_flag=True)
def init(overwrite_conftest, overwrite_pylenium_json, overwrite_pytest_ini):
    """ Initializes Pylenium into the current directory.

    By default, this creates (if they do not exist):
    * conftest.py
    * pylenium.json
    * pytest.ini

    If you want to overwrite these existing files, use the available Options.
    """
    scripts_path = os.path.dirname(os.path.abspath(__file__))
    user_cwd = os.getcwd()

    conftest = f'{scripts_path}/conftest.py'
    pylenium_json = f'{scripts_path}/pylenium.json'
    pytest_ini = f'{scripts_path}/pytest.ini'

    # initialize conftest.py
    if os.path.exists(f'{user_cwd}/conftest.py'):
        if overwrite_conftest:
            _copy(file=conftest, to_dir=user_cwd, message='conftest.py was overwritten at:')
        else:
            click.echo('conftest.py already exists at this location. '
                       'Use -c flag if you want to replace it with the latest.')
    else:
        _copy(file=conftest, to_dir=user_cwd, message='conftest.py was created at:')

    # initialize pylenium.json
    if os.path.exists(f'{user_cwd}/pylenium.json'):
        if overwrite_pylenium_json:
            _copy(file=pylenium_json, to_dir=user_cwd, message='pylenium.json was overwritten at:')
        else:
            click.echo('pylenium.json already exists at this location. '
                       'Use -p flag if you want to replace it with the latest defaults.')
    else:
        _copy(file=pylenium_json, to_dir=user_cwd, message='pylenium.json was created at:')

    # initialize pytest.ini
    if os.path.exists(f'{user_cwd}/pytest.ini'):
        if overwrite_pytest_ini:
            _copy(file=pytest_ini, to_dir=user_cwd, message='pytest.ini was overwritten at:')
        else:
            click.echo('pytest.ini already exists at this location. '
                       'Use -i flag if you want to replace it with the latest defaults.')
    else:
        _copy(file=pytest_ini, to_dir=user_cwd, message='pytest.ini was created at:')


@cli.command()
def joy():
    custom_fig = Figlet(font='colossal')
    click.echo(custom_fig.renderText('Pyl e n i u m Sparks Joy'))


# REPORT PORTAL #
#################

@cli.group()
def portal():
    """ CLI Commands to work with ReportPortal.io """
    pass


@portal.command()
def download():
    """ Download the ReportPortal docker-compose.yml file as docker-compose.report-portal.yml """
    report_portal.download_compose_yaml_file()
    click.echo('[SUCCESS] docker-compose.report-portal.yml file created!')
    click.echo('Depending on your OS, you will want to configure this file and your machine.')
    click.echo('Use the ReportPortal docs to see how you should configure your environment and .yml file:')
    click.echo('''
        ReportPortal Deploy with Docker
            https://reportportal.io/docs/Deploy-with-Docker
    ''')
    click.echo('Once configured, just use `$ pylenium portal up` to spin up your instance of ReportPortal!')


@portal.command()
def up():
    """ Spin up the ReportPortal instance using docker-compose.report-portal.yml """
    report_portal.compose_up()
    click.echo('[SUCCESS] ReportPortal instance created!')
    click.echo('1. Open the portal by going to: http://localhost:8080')
    click.echo('2. Login using the Admin or User credentials:')
    click.echo('''
        Admin
            username: superadmin
            password: erebus
        User
            username: default
            password: 1q2w3e
    ''')
    click.echo("* Make sure to change your password once you've logged in!")
    click.echo('3. Copy the ACCESS TOKEN, in User Profile, and paste it in the UUID variable in pytest.ini')
    click.echo('''
        All set! Now run your tests using `--reportportal`. For more info on the pytest plugin, visit:
            https://github.com/reportportal/agent-python-pytest
    ''')


@portal.command()
def down():
    """ Tear down the ReportPortal instance. """
    report_portal.down()
    click.echo('[SUCCESS] ReportPortal instance removed!')


if __name__ == '__main__':
    cli()



""" ReportPortal.io integration

1. Download the ReportPortal `docker-compose.yml` file as "docker-compose.report-portal.yml"
2. Setup permissions for ElasticSearch
3. Configure the `YAML` file based on OS
4. `docker-compose up`
5. Open ReportPortal and login (change password afterwards)
"""


import platform
from pylenium.scripts import cli_utils


def __stop_containers():
    """ Stop all ReportPortal containers.

    Returns:
        `CompletedProcess`
    """
    command = 'docker stop $(docker ps -a -f "name=reportportal" --format "{{.Names}}")'
    if platform.system() == 'Windows':
        command = "FOR /f \"tokens=*\" %i IN " \
                  "('docker ps -a -f \"name=reportportal\" --format \"{{.Names}}\"') " \
                  "DO docker stop %i"

    stop_containers_response = cli_utils.run_process(command, shell=True)
    if stop_containers_response.returncode != 0:
        raise EnvironmentError(f'[FAILED] {command}'
                               '\n\nUnable to stop ReportPortal containers:'
                               '\n * Make sure Docker is installed and running'
                               '\n * Make sure this command is run in the same dir as docker-compose.report-portal.yml'
                               f'\nResponse: {stop_containers_response}')
    return stop_containers_response


def __remove_containers():
    """ Remove all ReportPortal containers that are stopped.

    Returns:
        `CompletedProcess`
    """
    command = 'docker rm $(docker ps -a -f "name=reportportal" --format "{{.Names}}")'
    if platform.system() == 'Windows':
        command = "FOR /f \"tokens=*\" %i IN " \
                  "('docker ps -a -f \"name=reportportal\" --format \"{{.Names}}\"') " \
                  "DO docker rm %i"

    remove_containers_response = cli_utils.run_process(command, shell=True)
    if remove_containers_response.returncode != 0:
        raise EnvironmentError(f'[FAILED] {command}'
                               '\n\nUnable to remove ReportPortal containers after stopping them.'
                               f'\nResponse: {remove_containers_response}')
    return remove_containers_response


def download_compose_yaml_file():
    """ Download the ReportPortal docker-compose.yml file.

    * It is recommended to run this from the Project Root because
    this places the file as "docker-compose.report-portal.yml" in the context where this command was run.

    Returns:
        `CompletedProcess` if successful.

    Raises:
        `ConnectionError` if process returns non-zero status code.
    """
    response = cli_utils.run_process([
        'curl', 'https://raw.githubusercontent.com/reportportal/reportportal/master/docker-compose.yml',
        '-o', './docker-compose.report-portal.yml'
    ])
    if response.returncode != 0:
        raise ConnectionError(f'\n\nUnable to download docker-compose file from ReportPortal repo. '
                              f'\nResponse: {response}')
    return response


def compose_up():
    """ Spin up a ReportPortal instance using docker-compose.report-portal.yml.

    Returns:
        `CompletedProcess`

    Raises:
        `EnvironmentError` if process returns non-zero status code.
    """
    response = cli_utils.run_process([
        'docker-compose', '-p', 'reportportal',    # prefix containers with 'reportportal'
        '-f', 'docker-compose.report-portal.yml',  # use our auto-generated compose.yml
        'up', '-d', '--force-recreate'             # spin up in detached, "daemon mode"
    ])
    if response.returncode != 0:
        raise EnvironmentError('\n\nUnable to run "docker-compose" command to create ReportPortal instance.'
                               '\n * Make sure Docker is installed and running'
                               '\n * Make sure this command is run in the same dir as docker-compose.report-portal.yml'
                               f'\nResponse: {response}')
    return response


def down():
    """ Tear down the ReportPortal instance.

    This does not use the docker-compose.report-portal.yml file because, depending on Docker version, you may
    or may not have a network created that is not handled by docker-compose down.

    1. Stop all reportportal containers
    2. Kill (remove) all reportportal containers
    3. Remove the reportportal_default network (depends on docker version)

    Returns:
        `CompletedProcess` for the

    Raises:
        `EnvironmentError` if process returns non-zero status code.
    """
    __stop_containers()
    __remove_containers()
    remove_network_response = cli_utils.run_process([
        'docker', 'network', 'rm', 'reportportal_default'
    ])
    return remove_network_response



"""
`conftest.py` and `pylenium.json` files should stay at your Workspace Root.

conftest.py
    Although this file is editable, you should only change its contents if you know what you are doing.
    Instead, you can create your own conftest.py file in the folder where you store your tests.

pylenium.json
    You can change the values, but DO NOT touch the keys or you will break the schema.

py
    The only fixture you really need from this is `py`. This is the instance of Pylenium for each test.
    Just pass py into your test and you're ready to go!

Examples:
    def test_go_to_google(py):
        py.visit('https://google.com')
        assert 'Google' in py.title()
"""

import json
import logging
import os
import shutil
import sys

import pytest
import requests
from faker import Faker
from pytest_reportportal import RPLogger, RPLogHandler

from pylenium import Pylenium
from pylenium.config import PyleniumConfig, TestCase


def make_dir(filepath) -> bool:
    """ Make a directory.

    Returns:
        True if successful, False if not.
    """
    try:
        os.mkdir(filepath)
        return True
    except FileExistsError:
        return False


@pytest.fixture(scope='function')
def fake() -> Faker:
    """ A basic instance of Faker to make test data."""
    return Faker()


@pytest.fixture(scope='function')
def api():
    """ A basic instance of Requests to make HTTP API calls. """
    return requests


@pytest.fixture(scope="session")
def rp_logger(request):
    """ Report Portal Logger """
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    # Create handler for Report Portal if the service has been
    # configured and started.
    if hasattr(request.node.config, 'py_test_service'):
        # Import Report Portal logger and handler to the test module.
        logging.setLoggerClass(RPLogger)
        rp_handler = RPLogHandler(request.node.config.py_test_service)
        # Add additional handlers if it is necessary
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setLevel(logging.INFO)
        logger.addHandler(console_handler)
    else:
        rp_handler = logging.StreamHandler(sys.stdout)
    # Set INFO level for Report Portal handler.
    rp_handler.setLevel(logging.INFO)
    return logger


@pytest.fixture(scope='session', autouse=True)
def project_root() -> str:
    """ The Project (or Workspace) root as a filepath.

    * This conftest.py file should be in the Project Root if not already.
    """
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture(scope='session', autouse=True)
def test_run(project_root, request) -> str:
    """ Creates the `/test_results` directory to store the results of the Test Run.

    Returns:
        The `/test_results` directory as a filepath (str).
    """
    session = request.node
    test_results_dir = f'{project_root}/test_results'

    if os.path.exists(test_results_dir):
        # delete /test_results from previous Test Run
        shutil.rmtree(test_results_dir, ignore_errors=True)
    if not os.path.exists(test_results_dir):
        # create /test_results for this Test Run
        make_dir(test_results_dir)

    for test in session.items:
        # make the test_result directory for each test
        make_dir(f'{test_results_dir}/{test.name}')

    return test_results_dir


@pytest.fixture(scope='session')
def py_config(project_root, request) -> PyleniumConfig:
    """ Initialize a PyleniumConfig for each test

    1. This starts by deserializing the user-created pylenium.json from the Project Root.
    2. If that file is not found, then proceed with Pylenium Defaults.
    3. Then any CLI arguments override their respective key/values.
    """
    try:
        # 1. Load pylenium.json in Project Root, if available
        with open(f'{project_root}/pylenium.json') as file:
            _json = json.load(file)
        config = PyleniumConfig(**_json)
    except FileNotFoundError:
        # 2. pylenium.json not found, proceed with defaults
        config = PyleniumConfig()

    # 3. Override with any CLI args/options
    # Driver Settings
    cli_remote_url = request.config.getoption('--remote_url')
    if cli_remote_url:
        config.driver.remote_url = cli_remote_url

    cli_browser_options = request.config.getoption('--options')
    if cli_browser_options:
        config.driver.options = [option.strip() for option in cli_browser_options.split(',')]

    cli_browser = request.config.getoption('--browser')
    if cli_browser:
        config.driver.browser = cli_browser

    cli_capabilities = request.config.getoption('--caps')
    if cli_capabilities:
        # --caps must be in '{"name": "value", "boolean": true}' format
        # with double quotes around each key. booleans are lowercase.
        config.driver.capabilities = json.loads(cli_capabilities)

    cli_page_wait_time = request.config.getoption('--page_load_wait_time')
    if cli_page_wait_time and cli_page_wait_time.isdigit():
        config.driver.page_load_wait_time = int(cli_page_wait_time)

    # Logging Settings
    cli_pylog_level = request.config.getoption('--pylog_level')
    if cli_pylog_level:
        config.logging.pylog_level = cli_pylog_level

    cli_screenshots_on = request.config.getoption('--screenshots_on')
    if cli_screenshots_on:
        shots_on = True if cli_screenshots_on.lower() == 'true' else False
        config.logging.screenshots_on = shots_on

    cli_extensions = request.config.getoption('--extensions')
    if cli_extensions:
        config.driver.extension_paths = [ext.strip() for ext in cli_extensions.split(',')]

    return config


@pytest.fixture(scope='function')
def test_case(test_run, py_config, request) -> TestCase:
    """ Manages data pertaining to the currently running Test Function or Case.

        * Creates the test-specific logger.

    Args:
        test_run: The Test Run (or Session) this test is connected to.

    Returns:
        An instance of TestCase.
    """
    test_name = request.node.name
    test_result_path = f'{test_run}/{test_name}'
    py_config.driver.capabilities.update({'name': test_name})
    return TestCase(name=test_name, file_path=test_result_path)


@pytest.fixture(scope='function')
def py(test_case, py_config, request, rp_logger):
    """ Initialize a Pylenium driver for each test.

    Pass in this `py` fixture into the test function.

    Examples:
        def test_go_to_google(py):
            py.visit('https://google.com')
            assert 'Google' in py.title()
    """
    py = Pylenium(py_config)
    yield py
    try:
        if request.node.report.failed:
            # if the test failed, execute code in this block
            if py_config.logging.screenshots_on:
                screenshot = py.screenshot(f'{test_case.file_path}/test_failed.png')
                with open(screenshot, "rb") as image_file:
                    rp_logger.info("Test Failed - Attaching Screenshot",
                                   attachment={"name": "test_failed.png",
                                               "data": image_file,
                                               "mime": "image/png"})
    except AttributeError:
        rp_logger.error('Unable to access request.node.report.failed, unable to take screenshot.')
    except TypeError:
        rp_logger.info('Report Portal is not connected to this test run.')
    py.quit()


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """ Yield each test's outcome so we can handle it in other fixtures. """
    outcome = yield
    report = outcome.get_result()
    if report.when == 'call':
        setattr(item, "report", report)
    return report


def pytest_addoption(parser):
    parser.addoption(
        '--browser', action='store', default='', help='The lowercase browser name: chrome | firefox'
    )
    parser.addoption(
        '--remote_url', action='store', default='', help='Grid URL to connect tests to.'
    )
    parser.addoption(
        '--screenshots_on', action='store', default='', help="Should screenshots be saved? true | false"
    )
    parser.addoption(
        '--pylog_level', action='store', default='', help="Set the pylog_level: 'off' | 'info' | 'debug'"
    )
    parser.addoption(
        '--options', action='store',
        default='', help='Comma-separated list of Browser Options. Ex. "headless, incognito"'
    )
    parser.addoption(
        '--caps', action='store',
        default='', help='List of key-value pairs. Ex. \'{"name": "value", "boolean": true}\''
    )
    parser.addoption(
        '--page_load_wait_time', action='store',
        default='', help='The amount of time to wait for a page load before raising an error. Default is 0.'
    )
    parser.addoption(
        '--extensions', action='store',
        default='', help='Comma-separated list of extension paths. Ex. "*.crx, *.crx"'
    )



import logging
from typing import List, Union

import requests
from faker import Faker
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as ec

from pylenium import webdriver_factory, utils
from pylenium.config import PyleniumConfig
from pylenium.element import Element, Elements
from pylenium.switch_to import SwitchTo
from pylenium.wait import PyleniumWait


class PyleniumShould:
    def __init__(self, py: 'Pylenium', timeout: int, ignored_exceptions: list = None):
        self._py = py
        self._wait: PyleniumWait = self._py.wait(timeout=timeout, use_py=True, ignored_exceptions=ignored_exceptions)

    def have_title(self, title: str) -> 'Pylenium':
        """ An expectation that the title matches the given title.

        Args:
            title: The title to match.

        Returns:
            The current instance of Pylenium.

        Raises:
            `AssertionError` if the condition is not met within the timeout.
        """
        self._py.log.info('[ASSERT] .should().have_title()')
        try:
            value = self._wait.until(ec.title_is(title))
        except TimeoutException:
            value = False
        if value:
            return self._py
        else:
            self._py.log.critical('.should().have_title()')
            raise AssertionError(f'Expected Title: ``{title}`` - Actual Title: ``{self._py.title()}``')

    def contain_title(self, string: str) -> 'Pylenium':
        """ An expectation that the title contains the given string.

        Args:
            string: The case-sensitive string for the title to contain.

        Returns:
            The current instance of Pylenium.

        Raises:
            `AssertionError` if the condition is not met within the timeout.
        """
        self._py.log.info('[ASSERT] .should().contain_title()')
        try:
            value = self._wait.until(ec.title_contains(string))
        except TimeoutException:
            value = False
        if value:
            return self._py
        else:
            self._py.log.critical('.should().contain_title()')
            raise AssertionError(f'Expected ``{string}`` to be in ``{self._py.title()}``')

    def have_url(self, url: str) -> 'Pylenium':
        """ An expectation that the URL matches the given url.

        Args:
            url: The url to match.

        Returns:
            The current instance of Pylenium.

        Raises:
            `AssertionError` if the condition is not met within the timeout.
        """
        self._py.log.info('[ASSERT] .should().have_url()')
        try:
            value = self._wait.until(ec.url_to_be(url))
        except TimeoutException:
            value = False
        if value:
            return self._py
        else:
            self._py.log.critical('.should().have_url()')
            raise AssertionError(f'Expected URL: ``{url}`` - Actual URL: ``{self._py.url()}``')

    def contain_url(self, string: str) -> 'Pylenium':
        """ An expectation that the URL contains the given string.

        Args:
            string: The case-sensitive string for the url to contain.

        Returns:
            The current instance of Pylenium.

        Raises:
            `AssertionError` if the condition is not met within the timeout.
        """
        self._py.log.info('[ASSERT] .should().contain_url()')
        try:
            value = self._wait.until(ec.url_contains(string))
        except TimeoutException:
            value = False
        if value:
            return self._py
        else:
            self._py.log.critical('.should().contain_url()')
            raise AssertionError(f'Expected ``{string}`` to be in ``{self._py.url()}``')

    def not_find(self, css: str) -> bool:
        """ An expectation that there are no elements with the given CSS in the DOM.

        Args:
            css: The CSS selector.

        Returns:
            True if no elements are found.

        Raises:
            `AssertionError` if the condition is not met within the timeout.
        """
        self._py.log.info('[ASSERT] .should().not_find()')
        try:
            self._wait.until_not(lambda x: x.find_element(By.CSS_SELECTOR, css))
            return True
        except TimeoutException:
            self._py.log.critical('.should().not_find()')
            raise AssertionError(f'Found element with css: ``{css}``')

    def not_findx(self, xpath: str) -> bool:
        """ An expectation that there are no elements with the given XPATH in the DOM.

        Args:
            xpath: The XPATH selector.

        Returns:
            True if no elements are found.

        Raises:
            `AssertionError` if the condition is not met within the timeout.
        """
        self._py.log.info('[ASSERT] .should().not_find_xpath()')
        try:
            self._wait.until_not(lambda x: x.find_element(By.XPATH, xpath))
            return True
        except TimeoutException:
            self._py.log.critical('.should().not_findx()')
            raise AssertionError(f'Found element with xpath: ``{xpath}``')

    def not_contain(self, text: str) -> bool:
        """ An expectation that there are no elements with the given text in the DOM.

        Args:
            text: The text to contain.

        Returns:
            True if no elements are found.

        Raises:
            `AssertionError` if the condition is not met within the timeout.
        """
        self._py.log.info('[ASSERT] .should().not_contain()')
        try:
            self._wait.until_not(lambda x: x.find_element(By.XPATH, f'//*[contains(text(), "{text}")]'))
            return True
        except TimeoutException:
            self._py.log.critical('.should().not_contain()')
            raise AssertionError(f'Found element containing text: ``{text}``')


class Pylenium:
    """ The Pylenium API.

    V1.5.1+: Chrome is the default local browser

    Other supported browsers:
        * Firefox
        * Edge Chromium
        * IE
        * Opera
    """
    def __init__(self, config: PyleniumConfig):
        self.config = config
        self.log = logging.getLogger(__name__)
        self.fake = Faker()
        self.request = requests
        self._webdriver = None
        self._wait = None

    def init_webdriver(self):
        """ Initialize WebDriver using the Pylenium Config. """
        self._webdriver = webdriver_factory.build_from_config(self.config)
        caps = self._webdriver.capabilities
        try:
            self.log.info(f'Capabilities: '
                          f'browserName: {caps["browserName"]}, browserVersion: {caps["browserVersion"]}, '
                          f'platformName: {caps["platformName"]}, session_id: {self._webdriver.session_id}')
        except:
            self.log.warning(f'webdriver.capabilities did not have a key that Pylenium was expecting. '
                             f'Is your driver executable the right version?')

        # Default instance of PyleniumWait
        self._wait = PyleniumWait(self, self._webdriver, self.config.driver.wait_time, ignored_exceptions=None)

        # Initial Browser Setup
        if self.config.driver.page_load_wait_time:
            self.set_page_load_timeout(self.config.driver.page_load_wait_time)

        if self.config.viewport.maximize:
            self.maximize_window()
        else:
            self.viewport(self.config.viewport.width, self.config.viewport.height, self.config.viewport.orientation)
        return self._webdriver

    @property
    def webdriver(self) -> WebDriver:
        """ The current instance of Selenium's `WebDriver` API. """
        return self.init_webdriver() if self._webdriver is None else self._webdriver

    def title(self) -> str:
        """ The current page's title. """
        self.log.info('[STEP] py.title() - Get the current page title')
        return self.webdriver.title

    def url(self) -> str:
        """ The current page's URL. """
        self.log.info('[STEP] py.url() - Get the current page URL')
        return self.webdriver.current_url

    # NAVIGATION #
    ##############

    def visit(self, url: str) -> 'Pylenium':
        """ Navigate to the given URL.

        Returns:
            `py` so you can chain another command if needed.
        """
        self.log.info(f'[STEP] py.visit() - Visit URL: ``{url}``')
        self.webdriver.get(url)
        return self

    def go(self, direction: str, number: int = 1) -> 'Pylenium':
        """ Navigate forward or back.

        This command executes `window.history.go(number)`

        Args:
            direction: 'forward' or 'back'
            number: default is 1, will go back or forward one page in history.

        Examples:
            `py.go('back', 2)` will go back 2 pages in history.
            `py.go('forward')` will go forward 1 page in history.

        Returns:
            The current instance of Pylenium so you can chain commands.
        """
        self.log.info(f'[STEP] py.go() - Go {direction} {number} in browser history')
        if direction == 'back':
            self.webdriver.execute_script(f'window.history.go(arguments[0])', number * -1)
        elif direction == 'forward':
            self.webdriver.execute_script(f'window.history.go(arguments[0])', number)
        else:
            raise ValueError(f'direction was invalid. Must be `forward` or `back` but was {direction}')
        return self

    def reload(self) -> 'Pylenium':
        """ Reloads the current window. """
        self.log.info('[STEP] py.reload() - Reload the current page')
        self.webdriver.refresh()
        return self

    # FIND ELEMENTS #
    #################

    def contains(self, text: str, timeout: int = None) -> Element:
        """ Get the DOM element containing the `text`.

        * If timeout=None (default), use the default wait_time.
        * If timeout > 0, override the default wait_time.
        * If timeout=0, poll the DOM immediately without any waiting.

        Args:
            text: The text for the element to contain.
            timeout: The number of seconds to wait for this to succeed. Overrides the default wait_time.

        Returns:
            The first element that is found, even if multiple elements match the query.
        """
        self.log.info(f'[STEP] py.contains() - Find the element with text: ``{text}``')
        locator = (By.XPATH, f'//*[contains(text(), "{text}")]')

        if timeout == 0:
            element = self.webdriver.find_element(*locator)
        else:
            element = self.wait(timeout).until(
                lambda x: x.find_element(*locator),
                f'Could not find element with the text ``{text}``'
            )
        return Element(self, element, locator)

    def get(self, css: str, timeout: int = None) -> Element:
        """ Get the DOM element that matches the `css` selector.

        * If timeout=None (default), use the default wait_time.
        * If timeout > 0, override the default wait_time.
        * If timeout=0, poll the DOM immediately without any waiting.

        Args:
            css: The selector to use.
            timeout: The number of seconds to wait for this to succeed. Overrides the default wait_time.

        Returns:
            The first element that is found, even if multiple elements match the query.
        """
        self.log.info(f'[STEP] py.get() - Find the element with css: ``{css}``')
        by = By.CSS_SELECTOR

        if timeout == 0:
            element = self.webdriver.find_element(by, css)
        else:
            element = self.wait(timeout).until(
                lambda x: x.find_element(by, css),
                f'Could not find element with the CSS ``{css}``'
            )
        return Element(self, element, locator=(by, css))

    def find(self, css: str, timeout: int = None) -> Elements:
        """ Finds all DOM elements that match the `css` selector.

        * If timeout=None (default), use the default wait_time.
        * If timeout > 0, override the default wait_time.
        * If timeout=0, poll the DOM immediately without any waiting.

        Args:
            css: The selector to use.
            timeout: The number of seconds to wait for this to succeed. Overrides the default wait_time.

        Returns:
            A list of the found elements.
        """
        by = By.CSS_SELECTOR
        self.log.info(f'[STEP] py.find() - Find elements with css: ``{css}``')

        try:
            if timeout == 0:
                elements = self.webdriver.find_element(by, css)
            else:
                elements = self.wait(timeout).until(
                    lambda x: x.find_elements(by, css),
                    f'Could not find any elements with the CSS ``{css}``'
                )
        except TimeoutException:
            elements = []
        return Elements(self, elements, locator=(by, css))

    def getx(self, xpath: str, timeout: int = None) -> Element:
        """ Finds the DOM element that match the `xpath` selector.

        * If timeout=None (default), use the default wait_time.
        * If timeout > 0, override the default wait_time.
        * If timeout=0, poll the DOM immediately without any waiting.

        Args:
            xpath: The selector to use.
            timeout: The number of seconds to wait for this to succeed. Overrides the default wait_time.

        Returns:
            The first element that is found, even if multiple elements match the query.
        """
        by = By.XPATH
        self.log.info(f'[STEP] py.getx() - Find the element with xpath: ``{xpath}``')

        if timeout == 0:
            element = self.webdriver.find_element(by, xpath)
        else:
            element = self.wait(timeout).until(
                lambda x: x.find_element(by, xpath),
                f'Could not find an element with xpath: ``{xpath}``'
            )
        return Element(self, element, locator=(by, xpath))

    def findx(self, xpath: str, timeout: int = None) -> Elements:
        """ Finds the DOM elements that match the `xpath` selector.

        * If timeout=None (default), use the default wait_time.
        * If timeout > 0, override the default wait_time.
        * If timeout=0, poll the DOM immediately without any waiting.

        Args:
            xpath: The selector to use.
            timeout: The number of seconds to wait for this to succeed. Overrides the default wait_time.

        Returns:
            A list of the found elements.
        """
        by = By.XPATH
        self.log.info(f'[STEP] py.findx() - Find elements with xpath: ``{xpath}``')

        try:
            if timeout == 0:
                elements = self.webdriver.find_elements(by, xpath)
            else:
                elements = self.wait(timeout).until(
                    lambda x: x.find_elements(by, xpath),
                    f'Could not find an element with xpath: ``{xpath}``'
                )
        except TimeoutException:
            elements = []
        return Elements(self, elements, locator=(by, xpath))

    # EXPECTATIONS #
    ################

    def should(self, timeout: int = 0, ignored_exceptions: list = None) -> PyleniumShould:
        """ A collection of expectations for this driver.

        Examples:
            py.should().contain_title('QA at the Point')
            py.should().have_url('https://qap.dev')
        """
        if timeout:
            wait_time = timeout
        else:
            wait_time = self.config.driver.wait_time
        return PyleniumShould(self, wait_time, ignored_exceptions)

    # UTILITIES #
    #############

    def load_jquery(self, version: str = '3.5.1', timeout=None) -> 'Pylenium':
        """ Load jQuery onto the current page if it doesn't already exist.

        Args:
            version: The version of jQuery to load. Default version = 3.5.1
            timeout: The number of seconds to give this command to complete.

        Examples:
            py.load_jquery('3.5.1')
        """
        self.log.info('[STEP] .load_jquery() - Load jQuery onto the current page')
        jquery_url = f'https://code.jquery.com/jquery-{version}.min.js'
        load_jquery = utils.read_script_from_file('load_jquery.js')
        self.webdriver.execute_async_script(load_jquery, jquery_url)
        self.wait(timeout=timeout) \
            .until(lambda _: self.execute_script('return typeof(jQuery) !== "undefined";'),
                   message='jQuery was "undefined" which means it did not load within timeout.')
        return self

    def screenshot(self, filename: str) -> str:
        """ Take a screenshot of the current Window.

        Args:
            filename: the filepath including the filename and extension (like `.png`)

        Examples:
            py.screenshot('screenshots/home_page.png')
        """
        self.log.info(f'[STEP] py.screenshot() - Save screenshot to: {filename}')
        self.webdriver.save_screenshot(filename)
        return filename

    def scroll_to(self, x, y) -> 'Pylenium':
        """ Scroll to a location on the page.

        Args:
            x: The number of pixels to scroll horizontally.
            y: The number of pixels to scroll vertically.

        Examples:
            # Scroll down 500 px
            py.scroll_to(0, 500)
        """
        self.log.info(f'[STEP] py.scroll_to() - Scroll to ({x}, {y})')
        js = "window.scrollTo(arguments[0], arguments[1]);"
        self.webdriver.execute_script(js, x, y)
        return self

    def wait(self, timeout: int = None, use_py: bool = False, ignored_exceptions: list = None) -> Union[WebDriverWait, PyleniumWait]:
        """ The Wait object with the given timeout in seconds.

        If timeout=None or timeout=0,
        return the default instance of wait, else return a new instance of WebDriverWait or PyleniumWait.

        Args:
            timeout: The number of seconds to wait for the condition.
            use_py: True for a PyleniumWait.
            ignored_exceptions: List of exceptions for the condition to ignore.

        Default Ignored Exceptions:
            * NoSuchElementException

        Examples:
            # use the default wait_time in pylenium.json
            py.wait().until(lambda x: x.find_element_by_id('foo').get_attribute('style') == 'display: block;')
            # use a different timeout to control how long to wait for
            py.wait(5).until(lambda x: x.find_element_by_id('foo').is_displayed())
            py.wait(15, [NoSuchElementException, WebDriverException]).until(lambda x: x.find_element_by_id('foo'))
        """
        if timeout:  # if not None and greater than 0
            return self._wait.build(timeout, use_py, ignored_exceptions)
        else:
            return self._wait.build(self.config.driver.wait_time, use_py, ignored_exceptions)

    # BROWSER #
    ###########

    def delete_cookie(self, name):
        """ Deletes the cookie with the given name.

        Examples:
            `py.delete_cookie('cookie_name')`
        """
        self.log.info(f'py.delete_cookie() - Delete cookie named: {name}')
        self.webdriver.delete_cookie(name)

    def delete_all_cookies(self):
        """ Delete all cookies in the current session. """
        self.log.info('py.delete_all_cookies() - Delete all cookies')
        self.webdriver.delete_all_cookies()

    def get_cookie(self, name) -> dict:
        """ Get the cookie with the given name.

        Returns:
            The cookie if found, else None.

        Examples:
            py.get_cookie('cookie_name')
        """
        self.log.info(f'py.get_cookie() - Get cookie with name: {name}')
        return self.webdriver.get_cookie(name)

    def get_cookies(self):
        """ Get all cookies. """
        self.log.info('py.get_cookies() - Get all cookies')
        return self.webdriver.get_cookies()

    def set_cookie(self, cookie: dict):
        """ Adds a cookie to your current session.

        Args:
            cookie: A dictionary object, with required keys: "name" and "value";
                * optional keys: "path", "domain", "secure", "expiry"

        Examples:
            `py.set_cookie({'name' : 'foo', 'value' : 'bar'})`
            `py.set_cookie({'name' : 'foo', 'value' : 'bar', 'path' : '/'})`
            `py.set_cookie({'name' : 'foo', 'value' : 'bar', 'path' : '/', 'secure':True})`
        """
        self.log.info(f'py.set_cookie() - Set a cookie with name={cookie["name"]} and value={cookie["value"]}')
        self.webdriver.add_cookie(cookie)

    def execute_script(self, javascript: str, *args):
        """ Executes javascript in the current window or frame.

        Args:
            javascript: The script string to execute.
            args: Any arguments to be used in the script.

        Returns:
            The value returned by the script.

        Examples:
            `py.execute_script('return document.title;')`
            `py.execute_script('return document.getElementById(arguments[0]);', element_id)`
        """
        self.log.info('[STEP] py.execute_script() - Execute javascript into the Browser')
        return self.webdriver.execute_script(javascript, *args)

    def execute_async_script(self, javascript: str, *args):
        """ Executes javascript asynchronously in the current window or frame.

        Args:
            javascript: The script string to execute.
            args: Any arguments to be used in the script.

        Returns:
            The value returned by the script.

        Examples:
            `py.execute_async_script('return document.title;')`
        """
        self.log.info('[STEP] py.execute_async_script() - Execute javascript asynchronously into the Browser')
        return self.webdriver.execute_async_script(javascript, *args)

    def quit(self):
        """ Quits the driver.

        Closes any and every associated window.
        """
        self.log.info('py.quit() - Quit Pylenium and close all windows from the browser')
        self.webdriver.quit()

    # WINDOW #
    ##########

    @property
    def switch_to(self) -> SwitchTo:
        """ Switch between contexts like Windows or Frames. """
        return SwitchTo(self)

    @property
    def window_handles(self) -> List[str]:
        """ Returns the handles of all Windows in the current session. """
        return self.webdriver.window_handles

    @property
    def window_size(self) -> dict:
        """ Gets the width and height of the current Window. """
        return self.webdriver.get_window_size()

    def maximize_window(self) -> 'Pylenium':
        """ Maximizes the current Window. """
        self.log.info('py.maximize_window() - Maximize window')
        self.webdriver.maximize_window()
        return self

    def set_page_load_timeout(self, timeout: int) -> 'Pylenium':
        """ Set the amount of time to wait for a page load to complete before throwing an error.

        Args:
            timeout: The time to wait for.
        """
        self.webdriver.set_page_load_timeout(timeout)
        return self

    def viewport(self, width: int, height: int,  orientation: str = 'portrait') -> 'Pylenium':
        """ Control the size and orientation of the current context's browser window.

        Args:
            width: The width in pixels
            height: The height in pixels
            orientation: default is 'portrait'. Pass 'landscape' to reverse the width/height.

        Examples:
            py.viewport(1280, 800) # macbook-13 size
            py.viewport(1440, 900) # macbook-15 size
            py.viewport(375, 667)  # iPhone X size
        """
        self.log.info(f'[STEP] py.viewport() - Viewport set to width={width}, height={height}, orientation={orientation}')
        if orientation == 'portrait':
            self.webdriver.set_window_size(width, height)
        elif orientation == 'landscape':
            self.webdriver.set_window_size(height, width)
        else:
            raise ValueError(f'Orientation must be `portrait` or `landscape`.')
        return self



""" Factory to build WebDrivers. """
from typing import List, Optional

from selenium import webdriver
from selenium.webdriver.remote.webdriver import WebDriver
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.firefox import GeckoDriverManager
from webdriver_manager.microsoft import IEDriverManager, EdgeChromiumDriverManager
from webdriver_manager.opera import OperaDriverManager
from selenium.webdriver.edge.options import Options
from pylenium.config import PyleniumConfig


class Browser:
    """ ENUM of supported browsers. """
    CHROME = 'chrome'
    EDGE = 'edge'
    FIREFOX = 'firefox'
    IE = 'ie'
    OPERA = 'opera'


def build_capabilities(browser, capabilities: dict):
    """ Build the capabilities dictionary for WebDriver.

    Args:
        browser: The name of the browser.
        capabilities: The dict of capabilities to include.

    Examples:
        caps = WebDriverFactory().build_capabilities({'enableVNC': True, 'enableVideo': False})
    """
    caps = {}

    if browser == Browser.CHROME:
        caps.update(webdriver.DesiredCapabilities.CHROME.copy())
    elif browser == Browser.FIREFOX:
        caps.update(webdriver.DesiredCapabilities.FIREFOX.copy())
    elif browser == Browser.IE:
        caps.update(webdriver.DesiredCapabilities.INTERNETEXPLORER.copy())
    elif browser == Browser.OPERA:
        caps.update(webdriver.DesiredCapabilities.OPERA.copy())
    elif browser == Browser.EDGE:
        caps.update(webdriver.DesiredCapabilities.EDGE.copy())
    else:
        raise ValueError(f'{browser} is not supported.')

    if capabilities:
        caps.update(capabilities)

    return caps


def build_options(browser,
                  browser_options: List[str],
                  experimental_options: Optional[List[dict]],
                  extension_paths: Optional[List[str]]):
    """ Build the Options object for Chrome or Firefox.

    Args:
        browser: The name of the browser.
        browser_options: The list of options/arguments to include.
        experimental_options: The list of experimental options to include.
        extension_paths: The list of extensions to add to the browser session.

    Examples:
        WebDriverFactory().build_options('chrome', ['headless', 'incognito'], [{'useAutomationExtension', False}])
    """
    browser = browser.lower()
    if browser == Browser.CHROME:
        options = webdriver.ChromeOptions()
    elif browser == Browser.FIREFOX:
        options = webdriver.FirefoxOptions()
    elif browser == Browser.IE:
        options = webdriver.IeOptions()
    elif browser == Browser.OPERA:
        options = webdriver.ChromeOptions()
    elif browser == Browser.EDGE:
        options = Options()
    else:
        raise ValueError(f'{browser} is not supported. https://elsnoman.gitbook.io/pylenium/configuration/driver')

    for option in browser_options:
        options.add_argument(f'--{option}')

    if experimental_options:
        for exp_option in experimental_options:
            (name, value), = exp_option.items()
            options.add_experimental_option(name, value)

    if extension_paths:
        for path in extension_paths:
            options.add_extension(path)

    return options


def build_from_config(config: PyleniumConfig) -> WebDriver:
    """ Build a WebDriver using PyleniumConfig.

    PyleniumConfig is built using pylenium.json and CLI args.
    """
    if config.driver.remote_url:
        return build_remote(
            browser=config.driver.browser,
            remote_url=config.driver.remote_url,
            browser_options=config.driver.options,
            capabilities=config.driver.capabilities,
            experimental_options=config.driver.experimental_options,
            extension_paths=config.driver.extension_paths
        )
    browser = config.driver.browser.lower()
    if browser == Browser.CHROME:
        return build_chrome(config.driver.version, config.driver.options, config.driver.experimental_options, config.driver.extension_paths)
    elif browser == Browser.FIREFOX:
        return build_firefox(config.driver.version, config.driver.options, config.driver.experimental_options, config.driver.extension_paths)
    elif browser == Browser.IE:
        return build_ie(config.driver.version, config.driver.options, config.driver.capabilities, config.driver.experimental_options, config.driver.extension_paths)
    elif browser == Browser.OPERA:
        return build_opera(config.driver.version, config.driver.options, config.driver.experimental_options, config.driver.extension_paths)
    elif browser == Browser.EDGE:
        return build_edge(config.driver.version, config.driver.options, config.driver.capabilities, config.driver.experimental_options, config.driver.extension_paths)
    else:
        raise ValueError(f'{config.driver.browser} is not supported. https://elsnoman.gitbook.io/pylenium/configuration/driver')


def build_chrome(version: str,
                 browser_options: List[str],
                 experimental_options: Optional[List[dict]],
                 extension_paths: Optional[List[str]]) -> WebDriver:
    """ Build a ChromeDriver.

    Args:
        version: The desired version of Chrome.
        browser_options: The list of options/arguments to include.
        experimental_options: The list of experimental options to include.
        extension_paths: The list of extensions to add to the browser.

    Examples:
        driver = WebDriverFactory().build_chrome('latest', ['headless', 'incognito'], None)
    """
    options = build_options(Browser.CHROME, browser_options, experimental_options, extension_paths)
    return webdriver.Chrome(ChromeDriverManager(version=version).install(), options=options)


def build_edge(version: str,
               browser_options: List[str],
               capabilities: dict,
               experimental_options: Optional[List[dict]],
               extension_paths: Optional[List[str]]) -> WebDriver:
    """ Build a Edge Driver.

    Args:
        version: The desired version of Edge.
        browser_options: The list of options/arguments to include.
        capabilities: The dict of capabilities to include.
        experimental_options: The list of experimental options to include.
        extension_paths: The list of extensions to add to the browser.

    Examples:
        driver = WebDriverFactory().build_edge('latest', ['headless', 'incognito'], None)
    """
    caps = build_capabilities(Browser.EDGE, capabilities)
    options = build_options(Browser.EDGE, browser_options, experimental_options, extension_paths).to_capabilities()
    caps.update(options)
    return webdriver.Edge(EdgeChromiumDriverManager(version=version).install(),  capabilities=caps)


def build_firefox(version: str,
                  browser_options: List[str],
                  experimental_options: Optional[List[dict]],
                  extension_paths: Optional[List[str]]) -> WebDriver:
    """ Build a FirefoxDriver.

    Args:
        version: The desired version of Firefox.
        browser_options: The list of options/arguments to include.
        experimental_options: The list of experimental options to include.
        extension_paths: The list of extensions to add to the browser.

    Examples:
        driver = WebDriverFactory().build_firefox('latest', ['headless', 'incognito'], None)
    """
    options = build_options(Browser.FIREFOX, browser_options, experimental_options, extension_paths)
    return webdriver.Firefox(executable_path=GeckoDriverManager(version=version).install(), options=options)


def build_ie(version: str,
             browser_options: List[str],
             capabilities: dict,
             experimental_options: Optional[List[dict]],
             extension_paths: Optional[List[str]]) -> WebDriver:
    """ Build an IEDriver.

    Args:
        version: The desired version of IE.
        browser_options: The list of options/arguments to include.
        capabilities: The dict of capabilities.
        experimental_options: The list of experimental options to include.
        extension_paths: The list of extensions to add to the browser.

    Examples:
        driver = WebDriverFactory().build_ie('latest', ['headless'], None)
    """
    caps = build_capabilities(Browser.IE, capabilities)
    options = build_options(Browser.IE, browser_options, experimental_options, extension_paths)
    return webdriver.Ie(executable_path=IEDriverManager(version=version).install(), options=options, capabilities=caps)


def build_opera(version: str,
                browser_options: List[str],
                experimental_options: Optional[List[dict]],
                extension_paths: Optional[List[str]]) -> WebDriver:
    """ Build an OperaDriver.

    Args:
        version: The desired version of Opera.
        browser_options: The list of options/arguments to include.
        experimental_options: The list of experimental options to include.
        extension_paths: The list of extensions to add to the browser.

    Examples:
        driver = WebDriverFactory().build_opera('latest', ['--start-maximized'], None)
    """
    options = build_options(Browser.OPERA, browser_options, experimental_options, extension_paths)
    return webdriver.Opera(executable_path=OperaDriverManager(version=version).install(), options=options)


def build_remote(browser: str,
                 remote_url: str,
                 browser_options: List[str],
                 capabilities: dict,
                 experimental_options: Optional[List[dict]],
                 extension_paths: Optional[List[str]]) -> WebDriver:
    """ Build a RemoteDriver connected to a Grid.

    Args:
        browser: Name of the browser to connect to.
        remote_url: The URL to connect to the Grid.
        browser_options: The list of options/arguments to include.
        capabilities: The dict of capabilities to include.
        experimental_options: The list of experimental options to include.
        extension_paths: The list of extensions to add to the browser.

    Returns:
        The instance of WebDriver once the connection is successful
    """
    browser = browser.lower()
    caps = build_capabilities(browser, capabilities)
    options = build_options(browser, browser_options, experimental_options, extension_paths)

    return webdriver.Remote(
        command_executor=remote_url,
        desired_capabilities=caps,
        options=options
    )



import time
from typing import List, Tuple, Optional

from selenium.common.exceptions import NoSuchElementException, TimeoutException, WebDriverException
from selenium.webdriver import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebElement
from selenium.webdriver.support.select import Select
from selenium.webdriver.support import expected_conditions as ec

from pylenium import utils


class ElementWait:
    def __init__(self, webelement, timeout: int, ignored_exceptions: list = None):
        self._webelement = webelement
        self._timeout = 10 if timeout == 0 else timeout
        if ignored_exceptions:
            self._ignored_exceptions = ignored_exceptions
        else:
            self._ignored_exceptions = (
                NoSuchElementException
            )

    def until(self, method, message=''):
        screen = None
        stacktrace = None

        end_time = time.time() + self._timeout
        while True:
            try:
                value = method(self._webelement)
                if value:
                    return value
            except self._ignored_exceptions as exc:
                screen = getattr(exc, 'screen', None)
                stacktrace = getattr(exc, 'stacktrace', None)
            time.sleep(0.5)
            if time.time() > end_time:
                break
        raise TimeoutException(message, screen, stacktrace)


class ElementsShould:
    """ Expectations for the current list of elements, if any. """
    def __init__(self, py, elements: 'Elements', timeout: int, ignored_exceptions: list = None):
        self._py = py
        self._elements = elements
        self._wait = py.wait(timeout=timeout, use_py=True, ignored_exceptions=ignored_exceptions)

    def be_empty(self) -> bool:
        """ An expectation that the list has no elements.

        Returns:
            True if empty, else False.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info('  [ASSERT] .should().be_empty()')
        try:
            if self._elements.is_empty():
                return True
            else:
                locator = self._elements.locator
                value = self._wait.until(lambda drvr: len(drvr.find_elements(*locator)) == 0)
        except TimeoutException:
            value = False
        if value:
            return True
        else:
            self._py.log.critical('.should().be_empty()')
            raise AssertionError('List of elements was not empty.')

    def be_greater_than(self, length: int) -> bool:
        """ An expectation that the number of elements in the list is greater than the given length.

        Args:
            length: The length that the list should be greater than.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info(f'  [ASSERT] .should().be_greater_than() ``{length}``')
        try:
            if self._elements.length() > length:
                return True
            else:
                locator = self._elements.locator
                value = self._wait.until(lambda drvr: len(drvr.find_elements(*locator)) > length)
        except TimeoutException:
            value = False
        if value:
            return True
        else:
            self._py.log.critical('.should().be_greater_than()')
            raise AssertionError(f'Length of elements was not greater than {length}.')

    def be_less_than(self, length: int) -> bool:
        """ An expectation that the number of elements in the list is less than the given length.

        Args:
            length: The length that the list should be less than.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info(f'  [ASSERT] .should().be_less_than() ``{length}``')
        try:
            if self._elements.length() < length:
                return True
            else:
                locator = self._elements.locator
                value = self._wait.until(lambda drvr: len(drvr.find_elements(*locator)) < length)
        except TimeoutException:
            value = False
        if value:
            return True
        else:
            self._py.log.critical('.should().be_less_than()')
            raise AssertionError(f'Length of elements was not less than {length}.')

    def have_length(self, length: int) -> bool:
        """ An expectation that the number of elements in the list is equal to the given length.

        Args:
            length: The length that the list should be equal to.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info(f'  [ASSERT] .should().have_length() ``{length}``')
        try:
            if self._elements.length() == length:
                return True
            else:
                locator = self._elements.locator
                value = self._wait.until(lambda drvr: len(drvr.find_elements(*locator)) == length)
        except TimeoutException:
            value = False
        if value:
            return True
        else:
            self._py.log.critical('.should().have_length()')
            raise AssertionError(f'Length of elements was not equal to {length}.')

    def not_be_empty(self) -> 'Elements':
        """ An expectation that the list has at least one element.

        Returns:
            The list of elements if not empty.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info('  [ASSERT] .should().not_be_empty()')
        try:
            if not self._elements.is_empty():
                return self._elements
            else:
                locator = self._elements.locator
                value = self._wait.until(lambda drvr: drvr.find_elements(*locator))
        except TimeoutException:
            value = False
        if value:
            return Elements(self._py, value, self._elements.locator)
        else:
            self._py.log.critical('.should().not_be_empty()')
            raise AssertionError('List of elements was empty.')


class ElementShould:
    """ Expectations for the current element that is already in the DOM. """
    def __init__(self, py, element: 'Element', timeout: int, ignored_exceptions: list = None):
        self._py = py
        self._element = element
        self._wait = ElementWait(element.webelement, timeout, ignored_exceptions)

    # POSITIVE CONDITIONS #
    #######################

    def be_clickable(self) -> 'Element':
        """ An expectation that the element is displayed and enabled so you can click it.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info('  [ASSERT] .should().be_clickable()')
        try:
            value = self._wait.until(lambda e: e.is_displayed() and e.is_enabled())
        except TimeoutException:
            value = False
        if value:
            return self._element
        else:
            self._py.log.critical('.should().be_clickable()')
            raise AssertionError('Element was not clickable')

    def be_checked(self) -> 'Element':
        """ An expectation that the element is checked.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info('  [ASSERT] .should().be_checked()')
        try:
            value = self._wait.until(lambda e: self._element.is_checked())
        except TimeoutException:
            value = False
        if value:
            return self._element
        else:
            self._py.log.critical('.should().be_checked()')
            raise AssertionError('Element was not checked')

    def be_disabled(self) -> 'Element':
        """ An expectation that the element is disabled.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info('  [ASSERT] .should().be_disabled()')
        try:
            value = self._wait.until(lambda e: not e.is_enabled())
        except TimeoutException:
            value = False
        if value:
            return self._element
        else:
            self._py.log.critical('.should().be_disabled()')
            raise AssertionError('Element was not disabled')

    def be_enabled(self) -> 'Element':
        """ An expectation that the element is enabled.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info('  [ASSERT] .should().be_enabled()')
        try:
            value = self._wait.until(lambda e: e.is_enabled())
        except TimeoutException:
            value = False
        if value:
            return self._element
        else:
            self._py.log.critical('.should().be_enabled()')
            raise AssertionError('Element was not enabled')

    def be_focused(self) -> 'Element':
        """ An expectation that the element is focused.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info('  [ASSERT] .should().be_focused()')
        try:
            value = self._wait.until(lambda e: e == self._py.webdriver.switch_to.active_element)
        except TimeoutException:
            value = False

        if value:
            return self._element
        else:
            self._py.log.critical('.should().be_focused()')
            raise AssertionError('Element was not focused')

    def be_hidden(self) -> 'Element':
        """ An expectation that the element is not displayed but still in the DOM (aka hidden).

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info('  [ASSERT] .should().be_hidden()')
        try:
            value = self._wait.until(lambda e: e and not e.is_displayed())
        except TimeoutException:
            value = False

        if value:
            return self._element
        else:
            self._py.log.critical('.should().be_hidden()')
            raise AssertionError('Element was not hidden')

    def be_selected(self) -> 'Element':
        """ An expectation that the element is selected.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info('  [ASSERT] .should().be_selected()')
        try:
            value = self._wait.until(lambda e: e.is_selected())
        except TimeoutException:
            value = False

        if value:
            return self._element
        else:
            self._py.log.critical('.should().be_selected()')
            raise AssertionError('Element was not selected')

    def be_visible(self) -> 'Element':
        """ An expectation that the element is displayed.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info('  [ASSERT] .should().be_visible()')
        try:
            value = self._wait.until(lambda e: e and e.is_displayed())
        except TimeoutException:
            value = False

        if value:
            return self._element
        else:
            self._py.log.critical('.should().be_visible()')
            raise AssertionError('Element was not visible')

    def have_attr(self, attr: str, value: Optional[str] = None) -> 'Element':
        """ An expectation that the element has the given attribute with the given value.

        Args:
            attr: The name of the attribute.
            value (optional): The value of the attribute.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info(f'  [ASSERT] .should().have_attr() ``{attr}``')
        try:
            if value is None:
                val = self._wait.until(lambda e: e.get_attribute(attr))
            else:
                val = self._wait.until(lambda e: e.get_attribute(attr) == value)
        except TimeoutException:
            val = False

        if val:
            return self._element
        else:
            self._py.log.critical('.should().have_attr()')
            if value is None:
                raise AssertionError(f'Element did not have attribute: ``{attr}``')
            else:
                raise AssertionError(f'Expected Attribute Value: ``{value}`` '
                                     f'- Actual Attribute Value: ``{self._element.get_attribute("value")}``')

    def have_class(self, class_name: str) -> 'Element':
        """ An expectation that the element has the given className.

        Args:
            class_name: The `.className` of the element

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info(f'  [ASSERT] .should().have_class() ``{class_name}``')
        try:
            val = self._wait.until(lambda e: e.get_attribute('class') == class_name)
        except TimeoutException:
            val = False

        if val:
            return self._element
        else:
            self._py.log.critical('.should().have_class()')
            raise AssertionError(f'Expected className: ``{class_name}`` '
                                 f'- Actual className: ``{self._element.get_attribute("class")}``')

    def have_prop(self, prop: str, value: str) -> 'Element':
        """ An expectation that the element has the given property with the given value.

        Args:
            prop: The name of the property.
            value: The value of the property.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info(f'  [ASSERT] .should().have_prop() ``{prop}`` with value of ``{value}``')
        try:
            val = self._wait.until(lambda e: e.get_property(prop) == value)
        except TimeoutException:
            val = False

        if val:
            return self._element
        else:
            self._py.log.critical('.should().have_prop()')
            raise AssertionError(f'Expected Property value: ``{value}`` '
                                 f'- Actual Property value: ``{self._element.get_property(prop)}``')

    def have_text(self, text, case_sensitive=True) -> 'Element':
        """ An expectation that the element has the given text.

        Args:
            text: The exact text to match.
            case_sensitive: False if you want to ignore casing and leading/trailing spaces.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info(f'  [ASSERT] .should().have_text() ``{text}``')
        try:
            if case_sensitive:
                value = self._wait.until(lambda e: e.text == text)
            else:
                value = self._wait.until(lambda e: e.text.strip().lower() == text.lower())
        except TimeoutException:
            value = False

        if value:
            return self._element
        else:
            self._py.log.critical('.should().have_text()')
            raise AssertionError(f'Expected text: ``{text}`` - Actual text: ``{self._element.text()}``')

    def contain_text(self, text, case_sensitive=True) -> 'Element':
        """ An expectation that the element contains the given text.

        Args:
            text: The text that the element should contain.
            case_sensitive: False if you want to ignore casing and leading/trailing spaces.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info(f'  [ASSERT] .should().contain_text() ``{text}``')
        try:
            if case_sensitive:
                value = self._wait.until(lambda e: text in e.text)
            else:
                value = self._wait.until(lambda e: text.lower() in e.text.strip().lower())
        except TimeoutException:
            value = False

        if value:
            return self._element
        else:
            self._py.log.critical('.should().contain_text()')
            raise AssertionError(f'Expected ``{text}`` to be in ``{self._element.text()}``')

    def have_value(self, value) -> 'Element':
        """ An expectation that the element has the given value.

        Args:
            value: The exact value to match. Pass `None` if you expect the element not to have the value attribute.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.

        Examples:
            * An element with no `value` attribute will yield `None`
            * An element with a `value` attribute with no value will yield an empty string `""`
            * An element with a `value` attribute with a value will yield the value
        """
        self._py.log.info(f'  [ASSERT] .should().have_value() ``{value}``')
        try:
            val = self._wait.until(lambda e: e.get_attribute('value') == value)
        except TimeoutException:
            val = False

        if val:
            return self._element
        else:
            self._py.log.critical('.should().have_value()')
            raise AssertionError(f'Expected value: ``{value}`` '
                                 f'- Actual value: ``{self._element.get_attribute("value")}``')

    # NEGATIVE CONDITIONS #
    #######################

    def not_be_focused(self) -> 'Element':
        """ An expectation that the element is not focused.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info('  [ASSERT] .should().not_be_focused()')
        try:
            value = self._wait.until(lambda e: e != self._py.webdriver.switch_to.active_element)
        except TimeoutException:
            value = False

        if value:
            return self._element
        else:
            self._py.log.critical('.should().not_be_focused()')
            raise AssertionError('Element had focus')

    def disappear(self) -> 'Pylenium':
        """ An expectation that the element eventually disappears from the DOM.

        Returns:
            The current instance of Pylenium.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.

        Examples:
            # wait for a loading spinner to appear and then disappear once the load is complete
            py.get(#spinner).should().disappear()
        """
        self._py.log.info('  [ASSERT] .should().disappear()')
        try:
            value = self._wait.until(ec.invisibility_of_element(self._element.webelement))
        except TimeoutException:
            value = False
        if value:
            return self._py
        else:
            self._py.log.critical('.should().disappear()')
            raise AssertionError('Element was still visible or still in the DOM')

    def not_have_attr(self, attr: str, value: Optional[str] = None) -> 'Element':
        """ An expectation that the element does not have the given attribute with the given value.

        Either the attribute does not exist on the element or the value does not match the given value.

        Args:
            attr: The name of the attribute.
            value (optional): The value of the attribute.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info(f'  [ASSERT] .should().not_have_attr() ``{attr}``')
        try:
            if value is None:
                val = self._wait.until(lambda e: not e.get_attribute(attr))
            else:
                val = self._wait.until(lambda e: e.get_attribute(attr) is None or e.get_attribute(attr) != value)
        except TimeoutException:
            val = False

        if val:
            return self._element
        else:
            self._py.log.critical('.should().not_have_attr()')
            if value is None:
                raise AssertionError(f'Element had the attribute: ``{attr}``')
            else:
                raise AssertionError(f'Element still had attribute ``{attr}`` with the value of ``{value}``')

    def not_have_value(self, value) -> 'Element':
        """ An expectation that the element does not have the given value.

        Args:
            value: The exact value not to match. Pass `None` if you expect the element to have the value attribute.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.

        Examples:
            * An element with no `value` attribute will yield `None`
            * An element with a `value` attribute with no value will yield an empty string `""`
            * An element with a `value` attribute with a value will yield the value
        """
        self._py.log.info(f'  [ASSERT] .should().not_have_value() ``{value}``')
        try:
            val = self._wait.until(lambda e: e.get_attribute('value') != value)
        except TimeoutException:
            val = False

        if val:
            return self._element
        else:
            self._py.log.critical('.should().not_have_value()')
            raise AssertionError(f'Element had value matching ``{value}``')

    def not_have_text(self, text, case_sensitive=True) -> 'Element':
        """ An expectation that the element does not have the given text.

        Args:
            text: The exact text to match.
            case_sensitive: False if you want to ignore casing and leading/trailing spaces.

        Returns:
            The current element.

        Raises:
            `AssertionError` if the condition is not met in the specified amount of time.
        """
        self._py.log.info(f'  [ASSERT] .should().not_have_text() ``{text}``')
        try:
            if case_sensitive:
                value = self._wait.until(lambda e: e.text != text)
            else:
                value = self._wait.until(lambda e: e.text.strip().lower() != text.lower())
        except TimeoutException:
            value = False

        if value:
            return self._element
        else:
            self._py.log.critical('.should().not_have_text()')
            raise AssertionError(f'Element had the text matching ``{text}``')


class Elements(List['Element']):
    """ Represents a list of DOM elements. """
    def __init__(self, py, web_elements, locator: Optional[Tuple]):
        self._list = [Element(py, element, None) for element in web_elements]
        self._py = py
        self.locator = locator
        super().__init__(self._list)

    def should(self, timeout: int = 0, ignored_exceptions: list = None) -> ElementsShould:
        """ A collection of expectations for this list of elements.

        Examples:
            py.find('option').should().not_be_empty()
        """
        if timeout:
            wait_time = timeout
        else:
            wait_time = self._py.config.driver.wait_time
        return ElementsShould(self._py, self, wait_time, ignored_exceptions)

    def length(self) -> int:
        """ The number of elements in the list. """
        return len(self._list)

    def is_empty(self) -> bool:
        """ Checks if there are no elements in the list. """
        return self.length() == 0

    def first(self) -> 'Element':
        """ Gets the first element in the list.

        Raises:
            `IndexError` if the list is empty.
        """
        if self.length() > 0:
            return self._list[0]
        else:
            raise IndexError('Cannot get first() from an empty list.')

    def last(self) -> 'Element':
        """ Gets the last element in the list.

        Raises:
            `IndexError` if the list is empty.
        """
        if self.length() > 0:
            return self._list[-1]
        else:
            raise IndexError('Cannot get last() from an empty list.')

    # ACTIONS #
    ###########

    def check(self, allow_selected=False) -> 'Elements':
        """ Check all checkboxes or radio buttons in this list.

        Args:
            allow_selected: Do not raise error if any elements are already selected.

        Raises:
            `ValueError` if any elements are already selected.
            `ValueError` if any elements are not checkboxes or radio buttons.
        """
        self._py.log.info('  [STEP] elements.check() - Check all checkboxes or radio buttons in this list')
        for element in self._list:
            element.check(allow_selected)
        return self

    def uncheck(self, allow_deselected=False) -> 'Elements':
        """ Check all checkboxes or radio buttons in this list.

        Args:
            allow_deselected: Do not raise error if any elements are already deselected.

        Raises
            `ValueError` if any elements are already selected.
            `ValueError` if any elements are not checkboxes or radio buttons.
        """
        self._py.log.info('  [STEP] elements.uncheck() - Uncheck all checkboxes or radio buttons in this list')
        for element in self._list:
            element.uncheck(allow_deselected)
        return self

    # CONDITIONS #
    ##############

    def are_checked(self) -> bool:
        """ Check that all checkbox or radio buttons in this list are selected. """
        for element in self._list:
            if not element.is_checked():
                return False
        # every element is checked
        return True


class Element:
    """ Represents a DOM element. """
    def __init__(self, py, web_element: WebElement, locator: Optional[Tuple]):
        self._py = py
        self._webelement = web_element,
        self.locator = locator

    @property
    def webelement(self) -> WebElement:
        """ The current instance of the Selenium's `WebElement` API. """
        if isinstance(self._webelement, Tuple):
            return self._webelement[0]
        return self._webelement

    @property
    def py(self):
        """ The current instance of `py` that found this element. """
        return self._py

    def css_value(self, property_name: str):
        """ Gets the CSS Value of this element given the css_name. """
        self.py.log.info(f'  [STEP] .css_value() - Get a CSS Value given the property: ``{property_name}``')
        try:
            return self.webelement.value_of_css_property(property_name)
        except WebDriverException:
            self.py.log.warning(f'Property Name: ``{property_name}`` is invalid or not found.')
            return None

    def tag_name(self) -> str:
        """ Gets the tag name of this element. """
        tag_name = self.webelement.tag_name
        self.py.log.info(f'  [STEP] .tag_name() - Get the tag name of this element ``{tag_name}``')
        return tag_name

    def text(self) -> str:
        """ Gets the InnerText of this element. """
        text = self.webelement.text
        self.py.log.info(f'  [STEP] .text() - Get the text of this element ``{text}``')
        return text

    # EXPECTATIONS #
    ################

    def should(self, timeout: int = 0, ignored_exceptions: list = None) -> ElementShould:
        """ A collection of expectations for this element.

        Examples:
            py.get('#foo').should().be_clickable()
        """
        if timeout:
            wait_time = timeout
        else:
            wait_time = self._py.config.driver.wait_time
        return ElementShould(self.py, self, wait_time, ignored_exceptions)

    # METHODS #
    ###########

    def get_attribute(self, attribute: str):
        """ Gets the attribute's value.

            * If the value is 'true' or 'false', then this returns a Boolean
            * If the name does not exist, then return None
            * All other values are returned as strings

        Args:
            attribute: The name of the element's attribute.

        Returns:
            The value of the attribute. If the attribute does not exist, returns None
        """
        self.py.log.info(f'  [STEP] .get_attribute() - Get the ``{attribute}`` value of this element')
        value = self.webelement.get_attribute(attribute)
        if value == 'true':
            return True
        elif value == 'false':
            return False
        else:
            return value

    def get_property(self, prop: str):
        """ Gets the property's value.

        Args:
            prop: The name of the element's property.

        Returns:
            The value of the attribute.
        """
        self.py.log.info(f'  [STEP] .get_property() - Get the ``{prop}`` value of this element')
        return self.webelement.get_property(prop)

    # CONDITIONS #
    ##############

    def is_checked(self) -> bool:
        """ Check that this checkbox or radio button is selected.

        Raises:
            `ValueError` if element is not a checkbox or radio button
        """
        type_ = self.webelement.get_attribute('type')
        if type_ != 'checkbox' and type_ != 'radio':
            raise ValueError('Element is not a checkbox or radio button.')

        self.py.log.info(f'  [STEP] Check if this checkbox or radio button element is checked')
        return self.py.execute_script('return arguments[0].checked;', self.webelement)

    def is_displayed(self) -> bool:
        """ Check that this element is displayed.

        Raises:
            `NoSuchElementException` if the element is not in the DOM

        Returns:
            True if element is visible to the user, else False
        """
        self.py.log.info('  [STEP] Check if this element is displayed')
        return self.webelement.is_displayed()

    def is_enabled(self) -> bool:
        """ Check that this element is enabled.

        Returns:
            True if the element is enabled, else False
        """
        self.py.log.info('    [STEP] Check if this element is enabled')
        return self.webelement.is_enabled()

    def is_selected(self) -> bool:
        """ Check that this element is selected.

        Returns:
            True if the element is selected, else False
        """
        self.py.log.info('  [STEP] Check if this element is selected')
        return self.webelement.is_selected()

    # ACTIONS #
    ###########

    def check(self, allow_selected=False) -> 'Element':
        """ Check this checkbox or radio button.

        Args:
            allow_selected: Do not raise error if element is already selected.

        Raises:
            `ValueError` if element is already selected.
            `ValueError` if element is not a checkbox or radio button

        Returns:
            This element so you can chain another command.
        """
        self.py.log.info('  [STEP] .check() - Select this checkbox or radio button')
        type_ = self.webelement.get_attribute('type')
        if type_ == 'checkbox' or type_ == 'radio':
            checked = self.py.webdriver.execute_script('return arguments[0].checked;', self.webelement)
            if not checked:
                self.webelement.click()
                return self
            elif allow_selected:
                return self
            else:
                raise ValueError(f'{type_} is already selected.')
        raise ValueError('Element is not a checkbox or radio button.')

    def uncheck(self, allow_deselected=False) -> 'Element':
        """ Uncheck this checkbox or radio button.

        Args:
            allow_deselected: Do not raise error if element is already deselected.

        Raises:
            `ValueError` if element is already deselected.
            `ValueError` if element is not a checkbox or radio button

        Returns:
            This element so you can chain another command.
        """
        self.py.log.info('  [STEP] .uncheck() - Deselect this checkbox or radio button')
        type_ = self.webelement.get_attribute('type')
        if type_ == 'checkbox' or type_ == 'radio':
            checked = self.py.webdriver.execute_script('return arguments[0].checked;', self.webelement)
            if checked:
                self.webelement.click()
                return self
            elif allow_deselected:
                return self
            else:
                raise ValueError(f'{type_} is already deselected.')
        raise ValueError('Element is not a checkbox or radio button.')

    def clear(self) -> 'Element':
        """ Clears the text of the input or textarea element.

            * Only works on elements that can accept text entry.

        Returns:
            This element so you can chain another command.
        """
        self.py.log.info('  [STEP] .clear() - Clear the input of this element')
        self.webelement.clear()
        return self

    def click(self, force=False) -> 'Pylenium':
        """ Clicks the element.

        Args:
            force: If True, a JavascriptExecutor command is sent instead of Selenium's native `.click()`.

        Returns:
            The current instance of Pylenium so you can chain another command.
        """
        self.py.log.info('  [STEP] .click() - Click this element')
        if force:
            self.py.webdriver.execute_script('arguments[0].click()', self.webelement)
        else:
            self.webelement.click()
        return self.py

    def deselect(self, value) -> 'Pylenium':
        """ Deselects an `<option>` within a multi `<select>` element.

        Args:
            value: The value or text content of the `<option>` to be deselected.

        Raises:
            `ValueError` if this element is not a `<select>`

        Returns:
            The current instance of Pylenium so you can chain another command.
        """
        self.py.log.info('  [STEP] .deselect() - Deselect this element')
        if self.webelement.tag_name != 'select':
            raise ValueError(f'Can only perform Deselect on <select> elements. Tag name: {self.webelement.tag_name}')

        select = Select(self.webelement)
        if not select.is_multiple:
            raise NotImplementedError(f'Deselect can only be performed on multi-select elements.')

        try:
            select.deselect_by_visible_text(value)
        except NoSuchElementException:
            select.deselect_by_value(value)
        finally:
            return self.py

    def double_click(self) -> 'Pylenium':
        """ Double clicks the element.

        Returns:
            The current instance of Pylenium so you can chain another command.
        """
        self.py.log.info('  [STEP] .double_click() - Double click this element')
        ActionChains(self.py.webdriver).double_click(self.webelement).perform()
        return self.py

    def drag_to(self, css: str, jquery_version='3.5.1') -> 'Element':
        """ Drag the current element to another element given its CSS selector.

        Args:
            css: The CSS selector of the element to drag to.
            jquery_version: The version of jQuery to use for this action.

        Returns:
            The current element.

        Examples:
            py.get('#draggable-card').drag_to('#done-column')
        """
        self.py.log.info('  [STEP] .drag_to() - Drag this element to another element')
        to_element = self.py.get(css).webelement
        self.py.load_jquery(jquery_version)
        dnd_javascript = utils.read_script_from_file('drag_and_drop.js')
        self.py.execute_script(
            dnd_javascript + "$(arguments[0]).simulateDragDrop({ dropTarget: arguments[1] });",
            self.webelement,
            to_element
        )
        return self

    def drag_to_element(self, to_element: 'Element', jquery_version='3.5.1') -> 'Element':
        """ Drag the current element to the given element.

        Args:
            to_element: The Element to drag to.
            jquery_version: The version of jQuery to use for this action.

        Returns:
            The current element.

        Examples:
            column = py.get('#done-column')
            py.get('#draggable-card').drag_to_element(column)
        """
        self.py.log.info('  [STEP] .drag_to_element() - Drag this element to another element')
        self.py.load_jquery(jquery_version)
        dnd_javascript = utils.read_script_from_file('drag_and_drop.js')
        self.py.execute_script(
            dnd_javascript + "$(arguments[0]).simulateDragDrop({ dropTarget: arguments[1] });",
            self.webelement,
            to_element.webelement
        )
        return self

    def hover(self) -> 'Pylenium':
        """ Hovers the element.

        Returns:
            The current instance of Pylenium so you can chain another command.
        """
        self.py.log.info('  [STEP] .hover() - Hovers this element')
        ActionChains(self.py.webdriver).move_to_element(self.webelement).perform()
        return self.py

    def right_click(self) -> 'Pylenium':
        """ Right clicks the element.

        Returns:
            The current instance of Pylenium so you can chain another command.
        """
        self.py.log.info('  [STEP] .right_click() - Right click this element')
        ActionChains(self.py.webdriver).context_click(self.webelement).perform()
        return self.py

    def select(self, value) -> 'Element':
        """ Selects an `<option>` within a `<select>` element.

        Args:
            value: The value, text content or index of the `<option>` to be selected.

        Raises:
            `ValueError` if this element is not a `<select>`.

        Returns:
            This element so you can chain another command if needed.
        """
        self.py.log.info('  [STEP] .select() - Select an option in this element')
        if self.webelement.tag_name != 'select':
            raise ValueError(f'Can only perform Select on <select> elements. Current tag name: {self.webelement.tag_name}')
        try:
            Select(self.webelement).select_by_visible_text(value)
        except NoSuchElementException:
            Select(self.webelement).select_by_value(value)
        except BaseException:
            Select(self.webelement).select_by_index(value)
        finally:
            return self

    def select_many(self, values: list) -> 'Pylenium':
        """ Selects multiple `<options>` within a `<select>` element.

        Args:
            values: The list of values or text contents of the `<option>` to be selected.

        Raises:
            `ValueError` if this element is not a `<select>`.

        Returns:
            The current instance of Pylenium so you can chain another command.
        """
        self.py.log.info('  [STEP] .select_many() - Select many options in this element')
        if self.webelement.tag_name != 'select':
            raise ValueError(f'Can only perform Select on <select> elements. Current tag name: {self.webelement.tag_name}')

        select = Select(self.webelement)
        if not select.is_multiple:
            raise NotImplementedError(f'This <select> only allows a single option. Use .select() instead.')

        try:
            for val in values:
                select.select_by_visible_text(val)
        except NoSuchElementException:
            for val in values:
                select.select_by_value(val)
        finally:
            return self.py

    def submit(self) -> 'Pylenium':
        """ Submits the form.

            * Meant for <form> elements. May not have an effect on other types.

        Returns:
            The current instance of Pylenium so you can chain another command.
        """
        self.py.log.info('  [STEP] .submit() - Submit the form')
        self.webelement.submit()
        return self.py

    def type(self, *args) -> 'Element':
        """ Simulate a user typing keys into the input.

        Returns:
            This element so you can chain another command.
        """
        self.py.log.info('  [STEP] .type() - Type keys into this element')
        self.webelement.send_keys(args)
        return self

    # FIND ELEMENTS #
    #################

    def contains(self, text: str, timeout: int = None) -> 'Element':
        """ Gets the DOM element containing the `text`.

        Args:
            text: The text for the element to contain
            timeout: The number of seconds to find the element. Overrides default wait_time.

        Returns:
            The first element that is found, even if multiple elements match the query.
        """
        self.py.log.info(f'  [STEP] .contains() - Find the element that contains text: ``{text}``')
        locator = (By.XPATH, f'.//*[contains(text(), "{text}")]')

        if timeout == 0:
            element = self.webelement.find_element(*locator)
        else:
            element = self.py.wait(timeout).until(
                lambda _: self.webelement.find_element(*locator),
                f'Could not find element with the text ``{text}``'
            )
        return Element(self.py, element, locator)

    def get(self, css: str, timeout: int = None) -> 'Element':
        """ Gets the DOM element that matches the `css` selector in this element's context.

        Args:
            css: The selector
            timeout: The number of seconds to find the element. Overrides default wait_time.

        Returns:
            The first element that is found, even if multiple elements match the query.
        """
        self.py.log.info(f'  [STEP] .get() - Find the element that has css: ``{css}``')
        by = By.CSS_SELECTOR

        if timeout == 0:
            element = self.webelement.find_element(by, css)
        else:
            element = self.py.wait(timeout).until(
                lambda _: self.webelement.find_element(by, css),
                f'Could not find element with the CSS ``{css}``'
            )
        return Element(self.py, element, locator=(by, css))

    def find(self, css: str, timeout: int = None) -> Elements:
        """ Finds all DOM elements that match the `css` selector in this element's context.

        Args:
            css: The selector
            timeout: The number of seconds to find at least one element. Overrides default wait_time.

        Returns:
            A list of the found elements.
        """
        self.py.log.info(f'  [STEP] .find() - Find elements with css: ``{css}``')
        by = By.CSS_SELECTOR

        try:
            if timeout == 0:
                elements = self.webelement.find_elements(by, css)
            else:
                elements = self.py.wait(timeout).until(
                    lambda _: self.webelement.find_elements(by, css),
                    f'Could not find any elements with CSS ``{css}``'
                )
        except TimeoutException:
            elements = []
        return Elements(self.py, elements, locator=(by, css))

    def getx(self, xpath: str, timeout: int = None) -> 'Element':
        """ Finds the DOM element that matches the `xpath` selector.

        Args:
            xpath: The selector to use.
            timeout: The number of seconds to find the element. Overrides default wait_time.

        Returns:
            The first element found, even if multiple elements match the query.
        """
        self.py.log.info(f'  [STEP] .getx() - Find the element with xpath: ``{xpath}``')
        by = By.XPATH

        if timeout == 0:
            elements = self.webelement.find_element(by, xpath)
        else:
            elements = self.py.wait(timeout).until(
                lambda _: self.webelement.find_element(by, xpath),
                f'Could not find any elements with the xpath ``{xpath}``'
            )
        return Element(self, elements, locator=(by, xpath))

    def findx(self, xpath: str, timeout: int = None) -> 'Elements':
        """ Finds the DOM elements that matches the `xpath` selector.

        Args:
            xpath: The selector to use.
            timeout: The number of seconds to find at least one element. Overrides default wait_time.

        Returns:
            A list of the found elements.
        """
        self.py.log.info(f'  [STEP] .findx() - Find the elements with xpath: ``{xpath}``')
        by = By.XPATH

        try:
            if timeout == 0:
                elements = self.webelement.find_elements(by, xpath)
            else:
                elements = self.py.wait(timeout).until(
                    lambda _: self.webelement.find_elements(by, xpath),
                    f'Could not find any elements with the xpath ``{xpath}``'
                )
        except TimeoutException:
            elements = []
        return Elements(self, elements, locator=(by, xpath))

    def children(self) -> Elements:
        """ Gets the Child elements. """
        self.py.log.info('  [STEP] .children() - Get the children of this element')
        elements = self.py.webdriver.execute_script('return arguments[0].children;', self.webelement)
        return Elements(self.py, elements, None)

    def parent(self) -> 'Element':
        """ Gets the Parent element. """
        self.py.log.info('  [STEP] .parent() - Get the parent of this element')
        js = '''
        elem = arguments[0];
        return elem.parentNode;
        '''
        element = self.py.webdriver.execute_script(js, self.webelement)
        return Element(self.py, element, None)

    def siblings(self) -> Elements:
        """ Gets the Sibling elements. """
        self.py.log.info('  [STEP] .siblings() - Get the siblings of this element')
        js = '''
        elem = arguments[0];
        var siblings = [];
        var sibling = elem.parentNode.firstChild;

        while (sibling) {
            if (sibling.nodeType === 1 && sibling !== elem) {
                siblings.push(sibling);
            }
            sibling = sibling.nextSibling
        }
        return siblings;
        '''
        elements = self.py.webdriver.execute_script(js, self.webelement)
        return Elements(self.py, elements, None)

    # UTILITIES #
    #############

    def screenshot(self, filename) -> 'Element':
        """ Take a screenshot of the current element.

        Args:
            filename: the filepath including the filename and extension (like `.png`)

        Examples:
            py.get('#save-button').screenshot('elements/save-button.png')
        """
        self.py.log.info(f'  [STEP] .screenshot() - Take a screenshot and save to: ``{filename}``')
        self.webelement.screenshot(filename)
        return self

    def scroll_into_view(self) -> 'Element':
        """ Scroll this element into view. """
        self.py.log.info(f'  [STEP] .scroll_into_view() - Scroll this element into view')
        self.py.webdriver.execute_script('arguments[0].scrollIntoView(true);', self.webelement)
        return self

    def open_shadow_dom(self) -> 'Element':
        """ Open a Shadow DOM and return the Shadow Root element.

        Examples:
            shadow1 = py.get('extensions-manager').open_shadow_dom()
            shadow2 = py.get('extensions-list').open_shadow_dom()

        References:
            https://www.seleniumeasy.com/selenium-tutorials/accessing-shadow-dom-elements-with-webdriver
        """
        self.py.log.info('  [STEP] .open_shadow_dom() - Open a Shadow DOM and return the Root element')
        shadow_element = self.py.execute_script("return arguments[0].shadowRoot", self.webelement)
        return Element(self.py, shadow_element, locator=None)



import time
from typing import Tuple, Optional, Union

from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support.wait import WebDriverWait

from pylenium.element import Element, Elements


class PyleniumWait:
    """ The Pylenium version of Wait that returns Element and Elements objects."""
    def __init__(self, py, webdriver, timeout, ignored_exceptions: Optional[Tuple] = None):
        self._py = py
        self._webdriver = webdriver
        self._wait = WebDriverWait(
            driver=webdriver,
            timeout=timeout,
            ignored_exceptions=ignored_exceptions
        )

    def sleep(self, seconds: int):
        """ The test will sleep for the given number of seconds.

        Args:
            seconds: The number of seconds to sleep for.

        This is the same as using time.sleep()
        """
        time.sleep(seconds)

    def until(self, method, message=''):
        """ Wait until the method returns a non-False value.

        * The method uses Selenium WebDriver

        Args:
            method: The condition to wait for
            message: The message to show if the condition is not met in the time frame.

        Returns:
            If the value is a webelement, return a Pylenium Element object
            If the value is a list of WebElement, return a Pylenium Elements object
            Else return the non-False value

        Examples:
            # return an Element
            py.wait().until(lambda x: x.find_element_by_id('foo'), 'element "foo" was not found')
            # return Elements
            py.wait().until(lambda x: x.find_elements_by_xpath('//a'))
            # return True
            py.wait(5).until(lambda x: x.title  == 'QA at the Point')
        """
        value = self._wait.until(method, message)
        if isinstance(value, WebElement):
            return Element(self._py, value, None)
        if isinstance(value, list):
            try:
                return Elements(self._py, value, None)
            except:
                pass  # not a list of WebElement
        return value

    def until_not(self, method, message=''):
        """ Wait until the method returns a False value.

        * The method uses Selenium WebDriver

        Args:
            method: The condition to wait for
            message: The message to show if the condition is not met in the time frame.

        Returns:
            If the value is a webelement, return a Pylenium Element object
            If the value is a list of WebElement, return a Pylenium Elements object
            Else return the non-False value

        Examples:
            # wait until the title is not 'Home Page'
            py.wait().until_not(lambda x: x.title == 'Home Page' )
        """
        value = self._wait.until_not(method, message)
        if isinstance(value, WebElement):
            return Element(self._py, value, None)
        if isinstance(value, list):
            try:
                return Elements(self._py, value, None)
            except:
                pass  # not a list of WebElement
        return value

    def build(self, timeout: int, use_py=False, ignored_exceptions: list = None) -> Union[WebDriverWait, 'PyleniumWait']:
        """ Builds a WebDriverWait or PyleniumWait.

        Args:
            timeout: The number of seconds to wait for the condition to be True
            use_py: True if you want a PyleniumWait. False for a default WebDriverWait
            ignored_exceptions: List of exceptions to ignore in the Wait

        Returns:
            New instance of WebDriverWait if use_pylenium=False, else returns PyleniumWait
        """
        if use_py:
            return PyleniumWait(self._py, self._webdriver, timeout, ignored_exceptions)
        else:
            return WebDriverWait(self._webdriver, timeout, ignored_exceptions=ignored_exceptions)



import subprocess
from typing import List, Union


def run_process(tokenized_command: Union[List[str], str], shell=False) -> subprocess.CompletedProcess:
    """ Run a subprocess given the tokenized command.

    Args:
        tokenized_command: Tokenized list of strings that make up the command to run.
        shell: If the arg is a single string, change this to True

    Returns:
        `CompletedProcess` with - args, returncode, stderr, and stdout

    Raises:
        Does not raise a `CalledProcessError`, so make sure to check the returncode for a non-zero value.
    """
    response = subprocess.run(args=tokenized_command, stderr=subprocess.PIPE, stdout=subprocess.PIPE, shell=shell)
    return response



import pathlib


def read_script_from_file(file_name) -> str:
    """ Get the script string from a file in the scripts directory.

    Args:
        file_name: The file name with extension to read from.

    Examples:
        script = read_script_from_file('drag_and_drop.js')
    """
    path = str(pathlib.Path(__file__).parent.absolute())
    with open(path + f'/scripts/{file_name}', 'r') as file:
        script = file.read()
    return script



from setuptools import setup, find_packages
import pylenium as app


setup(
    name='pyleniumio',
    version=app.__version__,
    packages=find_packages(),
    include_package_data=True,
    url='https://github.com/ElSnoMan/pyleniumio',
    license='MIT',
    author='Carlos Kidman',
    author_email='carlos@qap.dev',
    description='The best of Selenium and Cypress in a single Python Package',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        'selenium', 'pytest', 'pytest-xdist', 'pytest-parallel', 'pydantic', 'pytest-reportportal',
        'faker', 'requests', 'webdriver-manager', 'click', 'pyfiglet'
    ],
    data_files=[('', [
        'pylenium/scripts/pylenium.json',
        'pylenium/scripts/pytest.ini',
        'pylenium/scripts/drag_and_drop.js',
        'pylenium/scripts/load_jquery.js'
    ])],
    entry_points='''
        [console_scripts]
        pylenium=pylenium.scripts.cli:cli
    '''
)



from selenium.common.exceptions import NoSuchFrameException
from selenium.webdriver.support import expected_conditions as ec


class SwitchTo:
    def __init__(self, pylenium):
        self._py = pylenium

    def frame(self, name_or_id: str, timeout: int = 0):
        """ Switch the driver's context to the new frame.

        Args:
            name_or_id: The frame's `id` or `name` attribute value
            timeout: The number of seconds to wait for the frame to be switched to.

        Examples:
            # Switch to an iframe
            py.switch_to.frame('main-frame')
        """
        self._py.log.info(f'[STEP] py.switch_to.frame() - Switch to frame using name or id: ``{name_or_id}``')
        self._py.wait(timeout).until(ec.frame_to_be_available_and_switch_to_it(name_or_id))
        return self._py

    def parent_frame(self):
        """ Switch the driver's context to the parent frame.

        If the parent frame is the current context, nothing happens.
        """
        self._py.log.info('[STEP] py.switch_to.parent_frame() - Switch to the parent frame')
        self._py.webdriver.switch_to.parent_frame()
        return self._py

    def default_content(self):
        """ Switch the driver's context to the default content. """
        self._py.log.info('[STEP] py.switch_to.default_content() - Switch to the default content of this browser session')
        self._py.webdriver.switch_to.default_content()
        return self._py

    def window(self, name_or_handle='', index=0):
        """ Switch the driver's context to the specified Window or Browser Tab.

        Args:
            name_or_handle: The name or window handle of the Window to switch to.
            index: The index position of the Window Handle.

        * `index=0` would be the default content.

        Examples:
            # Switch to a Window by handle
            windows = py.window_handles
            py.switch_to.window(name_or_handle=windows[1])

            # Switch to a newly opened Browser Tab by index
            py.switch_to.window(index=1)
        """
        if index:
            handle = self._py.webdriver.window_handles[index]
            self._py.log.info(f'[STEP] py.switch_to.window() - Switch to a Tab or Window by index: ``{index}``')
            self._py.webdriver.switch_to.window(handle)
            return self._py
        elif name_or_handle:
            self._py.log.info(f'[STEP] py.switch_to.window() - Switch to a Tab or Window by name or handle: ``{name_or_handle}``')
            self._py.webdriver.switch_to.window(name_or_handle)
            return self._py
        else:
            # context unchanged
            return self._py



from typing import List

from pydantic import BaseModel

# PYLENIUM CONFIG #
###################


class DriverConfig(BaseModel):
    browser: str = 'chrome'
    remote_url: str = ''
    wait_time: int = 10
    page_load_wait_time: int = 0
    options: List[str] = []
    capabilities: dict = {}
    experimental_options: List[dict] = None
    extension_paths: List[str] = None
    version: str = 'latest'


class LoggingConfig(BaseModel):
    pylog_level: str = 'info'
    screenshots_on: bool = True


class ViewportConfig(BaseModel):
    maximize: bool = True
    width: int = 1440
    height: int = 900
    orientation: str = 'portrait'


class PyleniumConfig(BaseModel):
    driver: DriverConfig = DriverConfig()
    logging: LoggingConfig = LoggingConfig()
    viewport: ViewportConfig = ViewportConfig()
    custom: dict = {}


# MODELS #
##########

class TestCase(BaseModel):
    name: str
    file_path: str

    class Config:
        arbitrary_types_allowed = True
